import dataclasses


# TODO: remove from dbt_common.contracts.util:: Replaceable + references
class Replaceable:
    def replace(self, **kwargs):
        return dataclasses.replace(self, **kwargs)


class Mergeable(Replaceable):
    def merged(self, *args):
        """Perform a shallow merge, where the last non-None write wins. This is
        intended to merge dataclasses that are a collection of optional values.
        """
        replacements = {}
        cls = type(self)
        for arg in args:
            for field in dataclasses.fields(cls):
                value = getattr(arg, field.name)
                if value is not None:
                    replacements[field.name] = value

        return self.replace(**replacements)



from contextvars import ContextVar, copy_context
from typing import List, Mapping, Optional

from dbt_common.constants import SECRET_ENV_PREFIX


class InvocationContext:
    def __init__(self, env: Mapping[str, str]):
        self._env = env
        self._env_secrets: Optional[List[str]] = None
        self.recorder = None
        # This class will also eventually manage the invocation_id, flags, event manager, etc.

    @property
    def env(self) -> Mapping[str, str]:
        return self._env

    @property
    def env_secrets(self) -> List[str]:
        if self._env_secrets is None:
            self._env_secrets = [
                v for k, v in self.env.items() if k.startswith(SECRET_ENV_PREFIX) and v.strip()
            ]
        return self._env_secrets


_INVOCATION_CONTEXT_VAR: ContextVar[InvocationContext] = ContextVar("DBT_INVOCATION_CONTEXT_VAR")


def reliably_get_invocation_var() -> ContextVar:
    invocation_var: Optional[ContextVar] = next(
        (cv for cv in copy_context() if cv.name == _INVOCATION_CONTEXT_VAR.name), None
    )

    if invocation_var is None:
        invocation_var = _INVOCATION_CONTEXT_VAR

    return invocation_var


def set_invocation_context(env: Mapping[str, str]) -> None:
    invocation_var = reliably_get_invocation_var()
    invocation_var.set(InvocationContext(env))


def get_invocation_context() -> InvocationContext:
    invocation_var = reliably_get_invocation_var()
    ctx = invocation_var.get()
    return ctx



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: types.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0btypes.proto\x12\x0bproto_types\x1a\x1fgoogle/protobuf/timestamp.proto\"\x91\x02\n\tEventInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x0b\n\x03msg\x18\x03 \x01(\t\x12\r\n\x05level\x18\x04 \x01(\t\x12\x15\n\rinvocation_id\x18\x05 \x01(\t\x12\x0b\n\x03pid\x18\x06 \x01(\x05\x12\x0e\n\x06thread\x18\x07 \x01(\t\x12&\n\x02ts\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x05\x65xtra\x18\t \x03(\x0b\x32!.proto_types.EventInfo.ExtraEntry\x12\x10\n\x08\x63\x61tegory\x18\n \x01(\t\x1a,\n\nExtraEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"6\n\x0eGenericMessage\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\"1\n\x11RetryExternalCall\x12\x0f\n\x07\x61ttempt\x18\x01 \x01(\x05\x12\x0b\n\x03max\x18\x02 \x01(\x05\"j\n\x14RetryExternalCallMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12,\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x1e.proto_types.RetryExternalCall\"#\n\x14RecordRetryException\x12\x0b\n\x03\x65xc\x18\x01 \x01(\t\"p\n\x17RecordRetryExceptionMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12/\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32!.proto_types.RecordRetryException\"@\n\x13SystemCouldNotWrite\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0b\n\x03\x65xc\x18\x03 \x01(\t\"n\n\x16SystemCouldNotWriteMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12.\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32 .proto_types.SystemCouldNotWrite\"!\n\x12SystemExecutingCmd\x12\x0b\n\x03\x63md\x18\x01 \x03(\t\"l\n\x15SystemExecutingCmdMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12-\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x1f.proto_types.SystemExecutingCmd\"\x1c\n\x0cSystemStdOut\x12\x0c\n\x04\x62msg\x18\x01 \x01(\t\"`\n\x0fSystemStdOutMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12\'\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x19.proto_types.SystemStdOut\"\x1c\n\x0cSystemStdErr\x12\x0c\n\x04\x62msg\x18\x01 \x01(\t\"`\n\x0fSystemStdErrMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12\'\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x19.proto_types.SystemStdErr\",\n\x16SystemReportReturnCode\x12\x12\n\nreturncode\x18\x01 \x01(\x05\"t\n\x19SystemReportReturnCodeMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12\x31\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32#.proto_types.SystemReportReturnCode\"\x19\n\nFormatting\x12\x0b\n\x03msg\x18\x01 \x01(\t\"\\\n\rFormattingMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12%\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x17.proto_types.Formatting\"\x13\n\x04Note\x12\x0b\n\x03msg\x18\x01 \x01(\t\"P\n\x07NoteMsg\x12$\n\x04info\x18\x01 \x01(\x0b\x32\x16.proto_types.EventInfo\x12\x1f\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x11.proto_types.Noteb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'types_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _EVENTINFO_EXTRAENTRY._options = None
  _EVENTINFO_EXTRAENTRY._serialized_options = b'8\001'
  _EVENTINFO._serialized_start=62
  _EVENTINFO._serialized_end=335
  _EVENTINFO_EXTRAENTRY._serialized_start=291
  _EVENTINFO_EXTRAENTRY._serialized_end=335
  _GENERICMESSAGE._serialized_start=337
  _GENERICMESSAGE._serialized_end=391
  _RETRYEXTERNALCALL._serialized_start=393
  _RETRYEXTERNALCALL._serialized_end=442
  _RETRYEXTERNALCALLMSG._serialized_start=444
  _RETRYEXTERNALCALLMSG._serialized_end=550
  _RECORDRETRYEXCEPTION._serialized_start=552
  _RECORDRETRYEXCEPTION._serialized_end=587
  _RECORDRETRYEXCEPTIONMSG._serialized_start=589
  _RECORDRETRYEXCEPTIONMSG._serialized_end=701
  _SYSTEMCOULDNOTWRITE._serialized_start=703
  _SYSTEMCOULDNOTWRITE._serialized_end=767
  _SYSTEMCOULDNOTWRITEMSG._serialized_start=769
  _SYSTEMCOULDNOTWRITEMSG._serialized_end=879
  _SYSTEMEXECUTINGCMD._serialized_start=881
  _SYSTEMEXECUTINGCMD._serialized_end=914
  _SYSTEMEXECUTINGCMDMSG._serialized_start=916
  _SYSTEMEXECUTINGCMDMSG._serialized_end=1024
  _SYSTEMSTDOUT._serialized_start=1026
  _SYSTEMSTDOUT._serialized_end=1054
  _SYSTEMSTDOUTMSG._serialized_start=1056
  _SYSTEMSTDOUTMSG._serialized_end=1152
  _SYSTEMSTDERR._serialized_start=1154
  _SYSTEMSTDERR._serialized_end=1182
  _SYSTEMSTDERRMSG._serialized_start=1184
  _SYSTEMSTDERRMSG._serialized_end=1280
  _SYSTEMREPORTRETURNCODE._serialized_start=1282
  _SYSTEMREPORTRETURNCODE._serialized_end=1326
  _SYSTEMREPORTRETURNCODEMSG._serialized_start=1328
  _SYSTEMREPORTRETURNCODEMSG._serialized_end=1444
  _FORMATTING._serialized_start=1446
  _FORMATTING._serialized_end=1471
  _FORMATTINGMSG._serialized_start=1473
  _FORMATTINGMSG._serialized_end=1565
  _NOTE._serialized_start=1567
  _NOTE._serialized_end=1586
  _NOTEMSG._serialized_start=1588
  _NOTEMSG._serialized_end=1668
# @@protoc_insertion_point(module_scope)



import typing
from typing import Any, Dict, Optional, Sequence, Tuple, Type

def get_type_origin(typ: Type) -> Type: ...
def get_generic_name(typ: Type, short: bool = ...) -> str: ...
def get_args(typ: Optional[Type]) -> Tuple[Type, ...]: ...
def get_literal_values(typ: Type) -> Tuple[Any, ...]: ...
def type_name(
    typ: Optional[Type],
    short: bool = ...,
    resolved_type_params: Optional[Dict[Type, Type]] = ...,
    is_type_origin: bool = ...,
    none_type_as_none: bool = ...,
) -> str: ...
def is_special_typing_primitive(typ: Any) -> bool: ...
def is_generic(typ: Type) -> bool: ...
def is_typed_dict(typ: Type) -> bool: ...
def is_named_tuple(typ: Type) -> bool: ...
def is_new_type(typ: Type) -> bool: ...
def is_union(typ: Type) -> bool: ...
def is_optional(typ: Type, resolved_type_params: Optional[Dict[Type, Type]] = ...) -> bool: ...
def is_annotated(typ: Type) -> bool: ...
def get_type_annotations(typ: Type) -> Sequence[Any]: ...
def is_literal(typ: Type) -> bool: ...
def not_none_type_arg(
    type_args: Tuple[Type, ...], resolved_type_params: Optional[Dict[Type, Type]] = ...
) -> Optional[Type]: ...
def is_type_var(typ: Type) -> bool: ...
def is_type_var_any(typ: Type) -> bool: ...
def is_class_var(typ: Type) -> bool: ...
def is_final(typ: Type) -> bool: ...
def is_init_var(typ: Type) -> bool: ...
def get_class_that_defines_method(method_name: str, cls: Type) -> Optional[Type]: ...
def get_class_that_defines_field(field_name: str, cls: Type) -> Optional[Type]: ...
def is_dataclass_dict_mixin(typ: Type) -> bool: ...
def is_dataclass_dict_mixin_subclass(typ: Type) -> bool: ...
def collect_type_params(typ: Type) -> Sequence[Type]: ...
def resolve_type_params(
    typ: Type, type_args: Sequence[Type] = ..., include_bases: bool = ...
) -> Dict[Type, Dict[Type, Type]]: ...
def substitute_type_params(typ: Type, substitutions: Dict[Type, Type]) -> Type: ...
def get_name_error_name(e: NameError) -> str: ...
def is_dialect_subclass(typ: Type) -> bool: ...
def is_self(typ: Type) -> bool: ...
def is_required(typ: Type) -> bool: ...
def is_not_required(typ: Type) -> bool: ...
def get_function_arg_annotation(
    function: typing.Callable[[Any], Any],
    arg_name: typing.Optional[str] = ...,
    arg_pos: typing.Optional[int] = ...,
) -> typing.Type: ...
def get_function_return_annotation(
    function: typing.Callable[[typing.Any], typing.Any]
) -> typing.Type: ...
def is_unpack(typ: Type) -> bool: ...
def is_type_var_tuple(typ: Type) -> bool: ...
def hash_type_args(type_args: typing.Iterable[typing.Type]) -> str: ...
def iter_all_subclasses(cls) -> typing.Iterator[Type]: ...



"""The record module provides a mechanism for recording dbt's interaction with
external systems during a command invocation, so that the command can be re-run
later with the recording 'replayed' to dbt.

If dbt behaves sufficiently deterministically, we will be able to use the
record/replay mechanism in several interesting test and debugging scenarios.
"""
import functools
import dataclasses
import json
import os
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Type

from dbt_common.context import get_invocation_context


class Record:
    """An instance of this abstract Record class represents a request made by dbt
    to an external process or the operating system. The 'params' are the arguments
    to the request, and the 'result' is what is returned."""

    params_cls: type
    result_cls: Optional[type]

    def __init__(self, params, result) -> None:
        self.params = params
        self.result = result

    def to_dict(self) -> Dict[str, Any]:
        return {
            "params": self.params._to_dict() if hasattr(self.params, "_to_dict") else dataclasses.asdict(self.params),  # type: ignore
            "result": self.result._to_dict() if hasattr(self.result, "_to_dict") else dataclasses.asdict(self.result) if self.result is not None else None,  # type: ignore
        }

    @classmethod
    def from_dict(cls, dct: Mapping) -> "Record":
        p = (
            cls.params_cls._from_dict(dct["params"])
            if hasattr(cls.params_cls, "_from_dict")
            else cls.params_cls(**dct["params"])
        )
        r = (
            cls.result_cls._from_dict(dct["result"])  # type: ignore
            if hasattr(cls.result_cls, "_from_dict")
            else cls.result_cls(**dct["result"])
            if cls.result_cls is not None
            else None
        )
        return cls(params=p, result=r)


class Diff:
    """Marker class for diffs?"""

    pass


class RecorderMode(Enum):
    RECORD = 1
    REPLAY = 2


class Recorder:
    _record_cls_by_name: Dict[str, Type] = {}
    _record_name_by_params_name: Dict[str, str] = {}

    def __init__(self, mode: RecorderMode, recording_path: Optional[str] = None) -> None:
        self.mode = mode
        self._records_by_type: Dict[str, List[Record]] = {}
        self._replay_diffs: List["Diff"] = []

        if recording_path is not None:
            self._records_by_type = self.load(recording_path)

    @classmethod
    def register_record_type(cls, rec_type) -> Any:
        cls._record_cls_by_name[rec_type.__name__] = rec_type
        cls._record_name_by_params_name[rec_type.params_cls.__name__] = rec_type.__name__
        return rec_type

    def add_record(self, record: Record) -> None:
        rec_cls_name = record.__class__.__name__  # type: ignore
        if rec_cls_name not in self._records_by_type:
            self._records_by_type[rec_cls_name] = []
        self._records_by_type[rec_cls_name].append(record)

    def pop_matching_record(self, params: Any) -> Optional[Record]:
        rec_type_name = self._record_name_by_params_name[type(params).__name__]
        records = self._records_by_type[rec_type_name]
        match: Optional[Record] = None
        for rec in records:
            if rec.params == params:
                match = rec
                records.remove(match)
                break

        return match

    def write(self, file_name) -> None:
        with open(file_name, "w") as file:
            json.dump(self._to_dict(), file)

    def _to_dict(self) -> Dict:
        dct: Dict[str, Any] = {}

        for record_type in self._records_by_type:
            record_list = [r.to_dict() for r in self._records_by_type[record_type]]
            dct[record_type] = record_list

        return dct

    @classmethod
    def load(cls, file_name: str) -> Dict[str, List[Record]]:
        with open(file_name) as file:
            loaded_dct = json.load(file)

        records_by_type: Dict[str, List[Record]] = {}

        for record_type_name in loaded_dct:
            record_cls = cls._record_cls_by_name[record_type_name]
            rec_list = []
            for record_dct in loaded_dct[record_type_name]:
                rec = record_cls.from_dict(record_dct)
                rec_list.append(rec)  # type: ignore
            records_by_type[record_type_name] = rec_list

        return records_by_type

    def expect_record(self, params: Any) -> Any:
        record = self.pop_matching_record(params)

        if record is None:
            raise Exception()

        result_tuple = dataclasses.astuple(record.result)
        return result_tuple[0] if len(result_tuple) == 1 else result_tuple

    def write_diffs(self, diff_file_name) -> None:
        json.dump(
            self._replay_diffs,
            open(diff_file_name, "w"),
        )

    def print_diffs(self) -> None:
        print(repr(self._replay_diffs))


def get_record_mode_from_env() -> Optional[RecorderMode]:
    replay_val = os.environ.get("DBT_REPLAY")
    if replay_val is not None and replay_val != "0" and replay_val.lower() != "false":
        return RecorderMode.REPLAY

    record_val = os.environ.get("DBT_RECORD")
    if record_val is not None and record_val != "0" and record_val.lower() != "false":
        return RecorderMode.RECORD

    return None


def record_function(record_type, method=False, tuple_result=False):
    def record_function_inner(func_to_record):
        # To avoid runtime overhead and other unpleasantness, we only apply the
        # record/replay decorator if a relevant env var is set.
        if get_record_mode_from_env() is None:
            return func_to_record

        @functools.wraps(func_to_record)
        def record_replay_wrapper(*args, **kwargs):
            recorder: Recorder = None
            try:
                recorder = get_invocation_context().recorder
            except LookupError:
                pass

            if recorder is None:
                return func_to_record(*args, **kwargs)

            # For methods, peel off the 'self' argument before calling the
            # params constructor.
            param_args = args[1:] if method else args

            params = record_type.params_cls(*param_args, **kwargs)

            include = True
            if hasattr(params, "_include"):
                include = params._include()

            if not include:
                return func_to_record(*args, **kwargs)

            if recorder.mode == RecorderMode.REPLAY:
                return recorder.expect_record(params)

            r = func_to_record(*args, **kwargs)
            result = (
                None
                if r is None or record_type.result_cls is None
                else record_type.result_cls(*r)
                if tuple_result
                else record_type.result_cls(r)
            )
            recorder.add_record(record_type(params=params, result=result))
            return r

        return record_replay_wrapper

    return record_function_inner
